@@ -3,18 +3,19 @@ import multiprocessing
 from pathlib import Path
 import subprocess
 import time
+import os
 
 import pandas as pd
 
 from vw_executor.artifacts import Output, Predictions, Model8, Model9, Model
 from vw_executor.pool import SeqPool, MultiThreadPool, Pool
 from vw_executor.loggers import MultiLogger, ILogger
-from vw_executor.handlers import MultiHandler
 from vw_executor.vw_cache import VwCache
-from vw_executor.handlers import HandlerBase, ProgressBars
+from vw_executor.handlers import MultiHandler, HandlerBase, ProgressBars
 from vw_executor.vw_opts import VwOpts, InteractiveGrid, VwOptsLike, GridLike
 
-from typing import Iterable, Optional, Union, Dict, Any, Type, List
+from typing import Callable, Iterable, Optional, Union, Dict, Any, Type, List, Generator
+from itertools import chain
 from abc import ABC, abstractmethod
 
 
@@ -48,35 +49,83 @@ class _VwBin(_VwCore):
     def __init__(self, path: Path):
         super().__init__(path)
 
-    def run(self, args: str) -> str:
+    def run(self, args: str, out_path: Path) -> str:
         command = f'{self.path} {args}'
+        stdout_file = open(out_path.parent / (out_path.name + '.out.txt'), 'w')
+        stderr_temp = out_path.parent / (out_path.name + '.pending')
+        stderr_file = open(stderr_temp, 'w')
+
         process = subprocess.Popen(
             command.split(),
             universal_newlines=True,
             encoding='utf-8',
-            stdout=subprocess.PIPE,
-            stderr=subprocess.PIPE
+            stdout=stdout_file,
+            stderr=stderr_file
         )
-        error = process.communicate()[1]
-        return error
+
+        returncode = process.wait()
+        stdout_file.close()
+        stderr_file.close()
+
+        os.replace(stderr_temp, out_path)
+
+        return []
 
 
-def _run_pyvw(args: str) -> Iterable[str]:
+def _run_pyvw(args: str, filename=None) -> Iterable[str]:
+    import traceback
+
     from vowpalwabbit import pyvw
-    execution = pyvw.vw(args, enable_logging=True)
+    try:
+        execution = pyvw.vw(args, enable_logging=True)
+    except Exception as e:
+        if filename is not None:
+            stderr_temp = filename.parent / (filename.name + '.pending')
+            with stderr_temp.open('w') as f:
+                f.write(str(e)+"\n\n")
+                traceback.print_exc(file=f)
+            return []
+        else:
+            raise e
     execution.finish()
-    return [l.rstrip() for l in execution.get_log()]
+    result = [l.rstrip() for l in execution.get_log()]
+    if filename is not None:
+        stderr_temp = filename.parent / (filename.name + '.pending')
+        _save(result, stderr_temp)
+        os.replace(stderr_temp, filename)
+        return []
+    else:
+        return result
 
 
 class _VwPy(_VwCore):
     def __init__(self):
         super().__init__(None)
 
-    def run(self, args: str) -> Iterable[str]:
+    def run(self, args: str, filename=None) -> Iterable[str]:
         from multiprocessing import Pool
         with Pool(1) as p:
-            return p.apply(_run_pyvw, [args])
+            return p.apply(_run_pyvw, [args], {"filename": filename})
 
+def symlink(source:Path, link_name:Path):
+    import os
+
+    if os.name == "nt":
+        def symlink_ms(source, link_name):
+            from subprocess import call
+            returncode = call(['cmd', '/C','mklink', link_name, source], shell=False, stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
+            if returncode != 0:
+                raise Exception(f"Make sure developer mode is enabled. Failed to create symlink {link_name} -> {source}.")
+        return symlink_ms(source, link_name)
+    else:
+        return os.symlink(source, link_name)
+
+def create_symlink_if_exists(source:Path, link_name):
+    if link_name.exists():
+        raise ValueError('Trying to overwrite existing symlink. Please consider changing the destination folder.')
+
+    if source.exists():
+        symlink(source, link_name)
 
 class Task:
     job: 'Job'
@@ -101,6 +150,7 @@ class Task:
                  input_folder: Path,
                  model_file: Optional[Path],
                  model_folder: Path,
+                 order_position: Optional[int] = 0,
                  no_run: bool = False):
         self.job = job
         self._logger = logger
@@ -109,12 +159,76 @@ class Task:
         self.status = ExecutionStatus.NotStarted
         self.model_file = model_file
         self.model_folder = model_folder
+        self._order_position = order_position
         self._no_run = no_run
-        self.args = self._prepare_args(self.job.cache)
+        self.args = str(self._prepare_args(self.job.cache))
         self.start_time = None
         self.end_time = None
+    
+    def create_human_readeable_symlink(
+        self,
+        base_dir: Optional[Union[str, Path]] = None,
+        translate_output: Dict[str, str] = {"-p": "predictions.txt", "-f": "final_regressor.vwmodel",
+        "--extra_metrics": "extra_metrics.json", "--invert_hash": "invert_hash.txt", "--readable_model": "readable_model.txt"},
+        create_symlink: Callable = create_symlink_if_exists) -> Path:
+        if self.input_file.parent == self.input_file:
+            raise ValueError("Input files cannot be on the root folder")
 
-    def _prepare_args(self, cache: VwCache) -> str:
+        def remove_argdash(arg: str):
+            if len(arg) > 1 and arg[0] == "-" and arg[1] == "-":
+                return arg[2:]
+            elif arg[0] == "-":
+                return arg[1:]
+            else:
+                return arg
+        
+        argdirname = ".".join([remove_argdash(arg) for arg in self.job.name.split()])
+
+        if not base_dir:
+            from datetime import datetime
+            base_dir = Path.cwd() / "_results" / datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
+        else:
+            base_dir = Path(base_dir)
+        
+        input_file_dir = self.input_file.parent.absolute()
+        input_file_name = str(self._order_position) + "th_file"
+
+        stderr_temp = self.stdout.path.parent / (self.stdout.path.name + '.pending')
+
+        if self.status == ExecutionStatus.Failed:
+            task_dir = base_dir / "ERRORS" / argdirname / input_file_dir.name / input_file_name
+        else:
+            task_dir = base_dir / argdirname / input_file_dir.name / input_file_name
+        task_dir.mkdir(parents=True, exist_ok=True)
+
+        create_symlink(self.stdout.path.absolute(), task_dir / "stdout.txt")
+        create_symlink(self.input_file.absolute(), task_dir / self.input_file.name)
+
+        stdout_file = self.stdout.path.parent / (self.stdout.path.name + '.out.txt')
+        create_symlink(stderr_temp.absolute(), task_dir / "ERROR_stdout.txt")
+
+        if stderr_temp.exists():
+            create_symlink(stdout_file.absolute(), task_dir / "more_ERROR_stdout.txt")
+        else:
+            create_symlink(stdout_file.absolute(), task_dir / "more_stdout.txt")
+
+        for output_arg, filename in self.outputs.items():
+            create_symlink(filename.absolute(), task_dir / translate_output[output_arg])
+
+        if self.model_file:
+            create_symlink(self.model_folder.joinpath(self.model_file).absolute(), task_dir / "input_regressor.vwmodel")
+        
+        vw_opts = self._prepare_args(self.job.cache)
+        for o in self.outputs.keys():
+            vw_opts[o] = translate_output[o] + ".repro"
+
+        cmd_repro_file = task_dir / "cmd_repro.txt"
+        with open(cmd_repro_file, "w") as f:
+            f.write(f"cwd: {str(Path.cwd().absolute())}\n")
+            f.write(f"vw args: {str(vw_opts)}\n")
+        return task_dir
+
+    def _prepare_args(self, cache: VwCache) -> VwOpts:
         opts = self.job.opts.copy()
         opts[self.job.input_mode] = self.input_file
 
@@ -134,13 +248,14 @@ class Task:
 
         opts[self.job.input_mode] = input_full
         opts = VwOpts(dict(opts, **self.outputs))
-        return str(opts)
+        return opts
 
     def _execute(self) -> Union[str, Iterable[str]]:
         self._logger.debug(f'Executing: {self.args}')
-        return self.job.core.run(self.args)
+        return self.job.core.run(self.args, self.stdout.path)
 
     def run(self, reset: bool) -> None:
+        self.status = ExecutionStatus.Running
         result_files = list(self.outputs.values()) + [self.stdout.path]
         not_exist = next((p for p in result_files if not p.exists()), None)
         self.start_time = time.time()
@@ -150,12 +265,19 @@ class Task:
             if self._no_run:
                 raise Exception('Result is not found, and execution is deprecated')
 
-            result = self._execute()
-            _save(result, self.stdout.path)
+            try:
+                result = self._execute()
+                assert result == []
+                self.status = ExecutionStatus.Success if self.stdout.loss is not None else ExecutionStatus.Failed
+            except:
+                self.status = ExecutionStatus.Failed
+                raise
+            finally:
+                self.end_time = time.time()
         else:
             self._logger.debug(f'Result of vw execution is found: {self.args}')
-        self.end_time = time.time()
-        self.status = ExecutionStatus.Success if self.stdout.loss is not None else ExecutionStatus.Failed
+            self.end_time = time.time()
+            self.status = ExecutionStatus.Success if self.stdout.loss is not None else ExecutionStatus.Failed
 
     def reset_stdout(self) -> None:
         self.stdout.path.unlink()
@@ -163,8 +285,11 @@ class Task:
     def _get_artifact(self, key: str, artifact_type: Type) -> Any:
         return artifact_type(self.outputs[key])
 
-    def predictions(self, key: str) -> Predictions:
-        return self._get_artifact(key, Predictions)
+    def predictions(self, key: str, problem = None) -> Union[Predictions, Generator[Dict, None, None]]:
+        if not problem:
+            return self._get_artifact(key, Predictions)
+        events = problem.fget(self._get_artifact(key, Predictions))
+        return map(lambda i_d: dict(i_d[1], **{'i': i_d[0]}), enumerate(events))
 
     def model8(self, key: str) -> Model8:
         return self._get_artifact(key, Model8)
@@ -196,7 +321,7 @@ class Job:
     core: _VwCore
     cache: VwCache
     _logger: MultiLogger
-    _handler: MultiHandler
+    _handler: HandlerBase
     _tasks: List[Task]
     opts: VwOpts
     name: str
@@ -211,7 +336,7 @@ class Job:
                  opts: VwOpts,
                  outputs: List[str],
                  input_mode: str,
-                 handler: MultiHandler,
+                 handler: HandlerBase,
                  logger: MultiLogger):
         self.core = vw
         self.cache = cache
@@ -232,14 +357,16 @@ class Job:
         for i, t in enumerate(self._tasks):
             self._logger.info(f'Starting task {i}...     File name: {t.input_file}')
             self._handler.on_task_start(self, i)
-            t.run(reset)
-            self._handler.on_task_finish(self, i)
-            self._logger.info(f'Task {i} is finished: {t.status}')
-            if t.status == ExecutionStatus.Failed:
-                self.failed = t
-                break
-            for p in t.outputs:
-                self.outputs[p].append(t.outputs[p])
+            try:
+                t.run(reset)
+            finally:
+                self._handler.on_task_finish(self, i)
+                self._logger.info(f'Task {i} is finished: {t.status}')
+                for p in t.outputs:
+                    self.outputs[p].append(t.outputs[p])
+                if t.status == ExecutionStatus.Failed:
+                    self.failed = t
+                    break
 
         self.status = self.failed.status if self.failed is not None else ExecutionStatus.Success
         self._logger.info(f'Job is finished: {self.status}')
@@ -272,6 +399,13 @@ class Job:
     def metrics(self) -> pd.DataFrame:
         return pd.DataFrame([t.metrics for t in self._tasks])
 
+    def predictions(self, key: str, problem) -> Generator[Dict, None, None]:
+        if not problem:
+            raise ValueError('Problem should be defined for job')
+        for i, task in enumerate(self):
+            for event in task.predictions(key, problem):
+                yield dict(event, **{'file': i})
+
 
 class TestJob(Job):
     def __init__(self,
@@ -283,11 +417,11 @@ class TestJob(Job):
                  outputs: List[str],
                  input_mode: str,
                  no_run: bool,
-                 handler: MultiHandler,
+                 handler: HandlerBase,
                  logger: MultiLogger):
         super().__init__(vw, cache, opts, outputs, input_mode, handler, logger)
-        for f in files:
-            self._tasks.append(Task(self, self._logger, f, input_dir, None, cache.path, no_run))
+        for i, f in enumerate(files):
+            self._tasks.append(Task(self, self._logger, f, input_dir, None, cache.path, order_position=i, no_run=no_run))
 
 
 class TrainJob(Job):
@@ -300,14 +434,14 @@ class TrainJob(Job):
                  outputs: List[str],
                  input_mode: str,
                  no_run: bool,
-                 handler: MultiHandler,
+                 handler: HandlerBase,
                  logger: MultiLogger):
         if '-f' not in outputs:
             outputs.append('-f')
         super().__init__(vw, cache, opts, outputs, input_mode, handler, logger)
         for i, f in enumerate(files):
             model = None if i == 0 else self._tasks[i - 1].outputs_relative['-f']
-            self._tasks.append(Task(self, self._logger, f, input_dir, model, cache.path, no_run))
+            self._tasks.append(Task(self, self._logger, f, input_dir, model, cache.path, order_position=i, no_run=no_run))
 
 
 def _assert_path_is_supported(path: Union[str, Path]) -> Path:
@@ -322,7 +456,7 @@ class Vw:
     logger: MultiLogger
     pool: Pool
     no_run: bool
-    handler: MultiHandler
+    handler: HandlerBase
     reset: bool
     last_job: Optional[Job]
 
@@ -332,14 +466,14 @@ class Vw:
                  procs: int = max(1, multiprocessing.cpu_count() // 2),
                  no_run: bool = False,
                  reset: bool = False,
-                 handlers: Union[HandlerBase, List[HandlerBase]] = ProgressBars(),
-                 loggers: Optional[List[ILogger]] = None):
+                 handler: Optional[HandlerBase] = ProgressBars(),
+                 logger: Optional[ILogger] = None):
         self._cache = VwCache(_assert_path_is_supported(cache_path))
         self._vw = _VwBin(path) if path is not None else _VwPy()
-        self.logger = MultiLogger(loggers or [])
+        self.logger = logger or MultiLogger([])
         self.pool = SeqPool() if procs == 1 else MultiThreadPool(procs)
         self.no_run = no_run
-        self.handler = MultiHandler(handlers if isinstance(handlers, list) else [handlers])
+        self.handler = handler or MultiHandler([])
         self.reset = reset
         self.last_job = None
 
@@ -349,15 +483,15 @@ class Vw:
               procs: Optional[int] = None,
               no_run: Optional[bool] = None,
               reset: Optional[bool] = None,
-              handlers: Optional[List[HandlerBase]] = None,
-              loggers: Optional[List[ILogger]] = None) -> 'Vw':
+              handler: Optional[HandlerBase] = None,
+              logger: Optional[ILogger] = None) -> 'Vw':
         return Vw(cache_path or self._cache.path,
                   path or self._vw.path,
                   procs or self.pool.procs,
                   no_run if no_run is not None else self.no_run,
                   reset if reset is not None else self.reset,
-                  handlers if handlers is not None else self.handler.handlers,
-                  loggers if loggers is not None else self.logger.loggers)
+                  handler or self.handler,
+                  logger or self.logger)
 
     def _run_impl(self,
                   inputs: List[Path],
@@ -465,7 +599,7 @@ class Vw:
         import matplotlib.pyplot as plt
 
         def _run_and_plot(**options):
-            self.last_job = self._with(handlers=[])._run(
+            self.last_job = self._with(handler=None)._run(
                 inputs, locals()['options'], outputs, input_mode, input_dir, job_type)
             ax.clear()
             fig.suptitle('Loss')
